<?php
/**
 * Mail Integration for Office 365 - Mail Class
 *
 * Captures overridden wp_mail() events, instigating a Microsoft
 * Graph API request using OAuth2 authentication rather than
 * the SMTP protocol
 *
 * @author            Edward Cross
 * @copyright         2020 Cross Connected
 * @license           GPL-2.0-or-later
 **/
namespace Classes\Mail_Integration_365;

use TheNetworg\OAuth2\Client\Provider\Azure;
use League\OAuth2\Client\Provider\Exception\IdentityProviderException;

if (!class_exists("Classes\Mail_Integration_365\Mail")) {
    class Mail
    { 
        /**
         * Custom send() function using captured wp_mail() event arguments
         * to send an email via the Microsoft Graph API using OAuth2 credentials
         *
         * @param   string          $to             The email address the email is to be sent to
         * @param   string          $subject        The email subject
         * @param   string          $message        The main content of the email
         * @param   string/array    $headers        The email headers
         * @param   array           $attachments    The email file attachments
         * @param   string          $content_type   The email content type
         * @return  bool            $bool           Whether the email contents were sent successfully.
         */
        public static function send($to, $subject, $message, $headers, $attachments, $content_type)
        {
            try {
                // Get main plugin options
                $options = get_option(Admin::$options_name);

                // Set up the OAuth2 provider using the Azure OAuth2 PHP Library
                $provider = new Azure([
                    "clientId" => $options[Admin::$options_name . "_client_id_field"],
                    "clientSecret" => $options[Admin::$options_name . "_client_secret_field"],
                    "redirectUri" => $options[Admin::$options_name . "_redirect_uri_field"],
                    "tenant" => $options[Admin::$options_name . "_tenant_id_field"],
					
                ]);
				
                // Set Azure endpoint version
                $provider->defaultEndPointVersion = Azure::ENDPOINT_VERSION_2_0;   				
				
				// Get the base Microsoft Graph URI
				$base_graph_uri = esc_url_raw($provider->getRootMicrosoftGraphUri(null));
				
				// Set provider scope
				$provider->scope = "openid profile email offline_access " . $base_graph_uri . "/User.Read " . $base_graph_uri . "/Mail.Send " . $base_graph_uri . "/Mail.ReadWrite";

                $access_token = get_option(Admin::$options_name . "_access_token");

                $cc = false;
                $bcc = false;
                $reply_to = false;
                
                // Create array if $headers is a string
                if(!is_array($headers)) {
                    $headers = explode("\n", str_replace("\r\n", "\n", $headers));
                }
                
                // Process email headers
                foreach ($headers as $header) {
                    if (strpos($header, ':') === false) {
                        continue;
                    }

                    list($header_name, $header_content) = explode(':', trim($header), 2);

                    $header_name    = trim($header_name);
                    $header_content = trim($header_content);
                    
                    // Override filtered content type if header value is present
                    if ('content-type' === strtolower($header_name)) {
                        if (strpos($header_content, ';') !== false) {
                            list($type, $charset) = explode(';', $header_content);
                            $content_type = trim($type);
                        } else {
                            $content_type = trim($header_content);
                        }
                    } elseif('cc' === strtolower($header_name)) {
                        $cc = $header_content;
                    } elseif('bcc' === strtolower($header_name)) {
                        $bcc = $header_content;
                    } elseif('reply-to' === strtolower($header_name)) {
                        $reply_to = $header_content;
                    }
                }

                // Set MS Graph content type
                if ($content_type == "text/plain") {
                    $graph_content_type = "text";
                } elseif ($content_type == "text/html") {
                    $graph_content_type = "html";
                } else {
                    $graph_content_type = "text";
                }

                // Create the email body object that can be serialised as JSON
                $body = array(
                    "message" => array(
                        "subject" => $subject,
                        "body" => array(
                            "contentType" => $graph_content_type,
                            "content" => $message,
                        ),
                        "toRecipients" => array(),
                        "ccRecipients" => array(),
                        "bccRecipients" => array(),
                        "replyTo" => array()
                    ),
                    "saveToSentItems" => "true", // Set to store emails in saved items by default
                );
                
                // Parse to address/es
                if (!is_array($to) && empty(explode(",", $to))) {
                    array_push($body["message"]["toRecipients"], array("emailAddress" => array("address" => $to)));
                } else {
                    if (!is_array($to)) {
                        $to = explode(",", $to);
                    }
                    
                    foreach ($to as $email) {
                        array_push($body["message"]["toRecipients"], array("emailAddress" => array("address" => $email)));
                    }
                }

                // Parse cc address/es
                if($cc) {
                    if (!is_array($cc) && empty(explode(",", $cc))) {
                        array_push($body["message"]["ccRecipients"], array("emailAddress" => array("address" => $cc)));
                    } elseif (!is_array($cc)) {
                        $cc = explode(",", $cc);
                                        
                        foreach ($cc as $email) {
                            array_push($body["message"]["ccRecipients"], array("emailAddress" => array("address" => $email)));
                        }
                    }
                }

                // Parse bcc address/es
                if($bcc) {                
                    if (!is_array($bcc) && empty(explode(",", $bcc))) {
                        array_push($body["message"]["bccRecipients"], array("emailAddress" => array("address" => $bcc)));
                    } elseif (!is_array($bcc)) {
                        $bcc = explode(",", $bcc);
                                            
                        foreach ($bcc as $email) {
                            array_push($body["message"]["bccRecipients"], array("emailAddress" => array("address" => $email)));
                        }
                    }
                }

                // Parse reply-to address/es
                if($reply_to) {
					if (!is_array($reply_to) && empty(explode(",", $reply_to))) {
                        array_push($body["message"]["replyTo"], array("emailAddress" => array("address" => $reply_to)));
                    } elseif (!is_array($reply_to)) {
                        $reply_to = explode(",", $reply_to);                        
                        
                        foreach ($reply_to as $email) {
                            array_push($body["message"]["replyTo"], array("emailAddress" => array("address" => $email)));
                        }
                    }
                }
                
                // Send mail from a specified account
                if (isset($options[Admin::$options_name . "_enable_send_as_field"])) {
					if($options[Admin::$options_name . "_enable_send_as_field"] === 'on' && isset($options[Admin::$options_name . "_send_as_email_field"])) {
						$body["message"]["from"] = array(
							"emailAddress" => array(
								"address" => $options[Admin::$options_name . "_send_as_email_field"]
							)
						);
					}
                }              
                 
                // Check if access token has expired and update plugin options with new access token.
                // The post() function used to instigate the API request below, does also check this by
                // default. However, it does not return the new access token to save.
                if ($access_token->hasExpired()) {
                    // Request a new access token using the refresh token embedded in $access_token
                    $access_token = $provider->getAccessToken("refresh_token", [
                        "refresh_token" => $access_token->getRefreshToken()
                    ]);

                    // Update the current saved access token with the new one
                    update_option(Admin::$options_name . "_access_token", $access_token);
                }
                
                // Handle email attachments
                if (!$attachments || $attachments == "") {
                    // Send the API request via the Azure OAuth2 PHP library"s post() function
                    $provider->post($base_graph_uri . "/v1.0/me/sendMail", $body, $access_token);
                } else {
                    $total_size = 0;
                    $body["message"]["attachments"] = array();
                    
                    // Check if total attachments are over 3MB in size
                    if (is_array($attachments)) {
                        foreach ($attachments as $key => $file) {
                            $handle = fopen($file, 'r');
                            $file_size = fileSize($file);
                            $total_size += $file_size;
                            $attachment = array(
                                '@odata.type' => '#microsoft.graph.fileAttachment',
                                'name' => basename($file),
                                'contentType' => mime_content_type($file),
                                'contentBytes' => base64_encode(fread($handle, $file_size)),
                            );
                            array_push($body["message"]["attachments"], $attachment);
                        }
                    } else {
                        $handle = fopen($file, 'r');
                        $file_size = fileSize($attachments);
                        $total_size += $file_size;
                        $attachment = array(
                            '@odata.type' => '#microsoft.graph.fileAttachment',
                            'name' => basename($file),
                            'contentType' => mime_content_type($file),
                            'contentBytes' => base64_encode(fread($handle, $file_size)),
                        );
                        array_push($body["message"]["attachments"], $attachment);
                    }

                    if ($total_size > 3145728) {
                        // Set message to null to stop wp_mail returning true
                        $to = "";
						
                        throw new IdentityProviderException('Attachments were too large to email via Mail Integration 365');
                    } else {
                        // Send the API request via the Azure OAuth2 PHP library"s post() function
                        $provider->post($base_graph_uri . "/v1.0/me/sendMail", $body, $access_token);
                    }
                }
            } catch (IdentityProviderException $e) {
                echo $e->getMessage();
                
				return false;
            } finally {
                return true;
            }
        }
    }
}
